#!/usr/bin/python3
# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import io
import os
import sys
import copy
import errno
import fcntl

import typing
from types import TracebackType
from typing import Any, Dict, IO, Optional, TextIO, Type, TypeVar, Union

# TODO: Remove this once we only support Python >= 3.11.
from typing_extensions import Self, TypeAlias

from ._libpathrs_cffi import lib as libpathrs_so

if typing.TYPE_CHECKING:
    # mypy apparently cannot handle the "ffi: cffi.api.FFI" definition in
    # _libpathrs_cffi/__init__.pyi so we need to explicitly reference the type
    # from cffi here.
    import cffi

    ffi = cffi.FFI()
    CString: TypeAlias = cffi.FFI.CData
    CBuffer: TypeAlias = cffi.FFI.CData
else:
    from ._libpathrs_cffi import ffi

    CString: TypeAlias = ffi.CData
    CBuffer: TypeAlias = ffi.CData


def _cstr(pystr: str) -> CString:
    return ffi.new("char[]", pystr.encode("utf8"))


def _pystr(cstr: CString) -> str:
    s = ffi.string(cstr)
    assert isinstance(s, bytes)  # typing
    return s.decode("utf8")


def _cbuffer(size: int) -> CBuffer:
    return ffi.new("char[%d]" % (size,))


def _is_pathrs_err(ret: int) -> bool:
    return ret < libpathrs_so.__PATHRS_MAX_ERR_VALUE


class PathrsError(Exception):
    """
    Represents a libpathrs error. All libpathrs errors have a description
    (PathrsError.message) and errors that were caused by an underlying OS error
    (or can be translated to an OS error) also include the errno value
    (PathrsError.errno).
    """

    message: str
    errno: Optional[int]
    strerror: Optional[str]

    def __init__(self, message: str, /, *, errno: Optional[int] = None):
        # Construct Exception.
        super().__init__(message)

        # Basic arguments.
        self.message = message
        self.errno = errno

        # Pre-format the errno.
        self.strerror = None
        if errno is not None:
            try:
                self.strerror = os.strerror(errno)
            except ValueError:
                self.strerror = str(errno)

    @classmethod
    def _fetch(cls, err_id: int, /) -> Optional[Self]:
        if err_id >= 0:
            return None

        err = libpathrs_so.pathrs_errorinfo(err_id)
        if err == ffi.NULL:  # type: ignore[comparison-overlap,unused-ignore] # TODO: Make this check nicer...
            return None

        description = _pystr(err.description)
        errno = err.saved_errno or None

        # TODO: Should we use ffi.gc()? mypy doesn't seem to like our types...
        libpathrs_so.pathrs_errorinfo_free(err)
        del err

        return cls(description, errno=errno)

    def __str__(self) -> str:
        if self.errno is None:
            return self.message
        else:
            return "%s (%s)" % (self.message, self.strerror)

    def __repr__(self) -> str:
        return "Error(%r, errno=%r)" % (self.message, self.errno)

    def pprint(self, out: TextIO = sys.stdout) -> None:
        "Pretty-print the error to the given @out file."
        # Basic error information.
        if self.errno is None:
            print("pathrs error:", file=out)
        else:
            print("pathrs error [%s]:" % (self.strerror,), file=out)
        print("  %s" % (self.message,), file=out)


INTERNAL_ERROR = PathrsError("tried to fetch libpathrs error but no error found")


class FilenoFile(typing.Protocol):
    def fileno(self) -> int: ...


FileLike = Union[FilenoFile, int]


def _fileno(file: FileLike) -> int:
    if isinstance(file, int):
        # file is a plain fd
        return file
    else:
        # Assume there is a fileno method.
        return file.fileno()


def _clonefile(file: FileLike) -> int:
    return fcntl.fcntl(_fileno(file), fcntl.F_DUPFD_CLOEXEC)


# TODO: Switch to def foo[T](...): ... syntax with Python >= 3.12.
Fd = TypeVar("Fd", bound="WrappedFd")


class WrappedFd(object):
    """
    Represents a file descriptor that allows for manual lifetime management,
    unlike os.fdopen() which are tracked by the GC with no way of "leaking" the
    file descriptor for FFI purposes.

    pathrs will return WrappedFds for most operations that return an fd.
    """

    _fd: Optional[int]

    def __init__(self, file: FileLike, /):
        """
        Construct a WrappedFd from any file-like object.

        For most cases, the WrappedFd will take ownership of the lifetime of
        the file handle. This means you should  So a raw file descriptor must
        only be turned into a WrappedFd *once* (unless you make sure to use
        WrappedFd.leak() to ensure there is only ever one owner of the handle
        at a given time).

        However, for os.fdopen() (or similar Pythonic file objects that are
        tracked by the GC), we have to create a clone and so the WrappedFd is a
        copy.
        """
        # TODO: Maybe we should always clone to make these semantics less
        # confusing...?
        fd = _fileno(file)
        if isinstance(file, io.IOBase):
            # If this is a regular open file, we need to make a copy because
            # you cannot leak files and so the GC might close it from
            # underneath us.
            fd = _clonefile(fd)
        self._fd = fd

    def fileno(self) -> int:
        """
        Return the file descriptor number of this WrappedFd.

        Note that the file can still be garbage collected by Python after this
        call, so the file descriptor number might become invalid (or worse, be
        reused for an unrelated file).

        If you want to convert a WrappedFd to a file descriptor number and stop
        the GC from the closing the file, use WrappedFd.into_raw_fd().
        """
        if self._fd is None:
            raise OSError(errno.EBADF, "Closed file descriptor")
        return self._fd

    def leak(self) -> None:
        """
        Clears this WrappedFd without closing the underlying file, to stop GC
        from closing the file.

        Note that after this operation, all operations on this WrappedFd will
        return an error. If you want to get the underlying file handle and then
        leak the WrappedFd, just use WrappedFd.into_raw_fd() which does both
        for you.
        """
        if self._fd is not None and self._fd >= 0:
            self._fd = None

    def fdopen(self, mode: str = "r") -> IO[Any]:
        """
        Convert this WrappedFd into an os.fileopen() handle.

        This operation implicitly calls WrappedFd.leak(), so the WrappedFd will
        no longer be useful and you should instead use the returned os.fdopen()
        handle.
        """
        fd = self.fileno()
        try:
            file = os.fdopen(fd, mode)
            self.leak()
            return file
        except:
            # "Unleak" the file if there was an error.
            self._fd = fd
            raise

    @classmethod
    def from_raw_fd(cls: Type[Fd], fd: int, /) -> Fd:
        "Shorthand for WrappedFd(fd)."
        return cls(fd)

    @classmethod
    def from_file(cls: Type[Fd], file: FileLike, /) -> Fd:
        "Shorthand for WrappedFd(file)."
        return cls(file)

    def into_raw_fd(self) -> int:
        """
        Convert this WrappedFd into a raw file descriptor that GC won't touch.

        This is just shorthand for WrappedFd.fileno() to get the fileno,
        followed by WrappedFd.leak().
        """
        fd = self.fileno()
        self.leak()
        return fd

    def isclosed(self) -> bool:
        """
        Returns whether the underlying file descriptor is closed or the
        WrappedFd has been leaked.
        """
        return self._fd is None

    def close(self) -> None:
        """
        Manually close the underlying file descriptor for this WrappedFd.

        WrappedFds are garbage collected, so this is usually unnecessary unless
        you really care about the point where a file is closed.
        """
        if not self.isclosed():
            assert self._fd is not None  # typing
            if self._fd >= 0:
                os.close(self._fd)
                self._fd = None

    def clone(self) -> Self:
        "Create a clone of this WrappedFd that has a separate lifetime."
        if self.isclosed():
            raise ValueError("cannot clone closed file")
        assert self._fd is not None  # typing
        newfd = self._fd
        if self._fd >= 0:
            newfd = _clonefile(self._fd)
        return self.__class__.from_raw_fd(newfd)

    def __copy__(self) -> Self:
        "Identical to WrappedFd.clone()"
        # A "shallow copy" of a file is the same as a deep copy.
        return copy.deepcopy(self)

    def __deepcopy__(self, memo: Dict[int, Any]) -> Self:
        "Identical to WrappedFd.clone()"
        return self.clone()

    def __del__(self) -> None:
        "Identical to WrappedFd.close()"
        self.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        self.close()


# XXX: This is _super_ ugly but so is the one in CPython.
def _convert_mode(mode: str) -> int:
    mode_set = set(mode)
    flags = os.O_CLOEXEC

    # We don't support O_CREAT or O_EXCL with libpathrs -- use creat().
    if "x" in mode_set:
        raise ValueError("pathrs doesn't support mode='x', use Root.creat()")
    # Basic sanity-check to make sure we don't accept garbage modes.
    if len(mode_set & {"r", "w", "a"}) > 1:
        raise ValueError("must have exactly one of read/write/append mode")

    read = False
    write = False

    if "+" in mode_set:
        read = True
        write = True
    if "r" in mode_set:
        read = True
    if "w" in mode_set:
        write = True
        flags |= os.O_TRUNC
    if "a" in mode_set:
        write = True
        flags |= os.O_APPEND

    if read and write:
        flags |= os.O_RDWR
    elif write:
        flags |= os.O_WRONLY
    else:
        flags |= os.O_RDONLY

    # We don't care about "b" or "t" since that's just a Python thing.
    return flags


class SingletonClass(type):
    """Metaclass used to create singleton classes."""

    _instances: dict[type, Type[Any]] = {}

    def __call__(cls, *args, **kwargs):  # type: ignore[no-untyped-def] # TODO: Not clear what annotations to use, and mypy appears to be confused by metaclasses.
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonClass, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



#!/usr/bin/python3
# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import os

import typing
from typing import Any, IO, Union, cast
import warnings

# TODO: Remove this once we only support Python >= 3.11.
from typing_extensions import TypeAlias

from ._internal import (
    # Generic helpers.
    SingletonClass,
    # File type helpers.
    FileLike,
    WrappedFd,
    _convert_mode,
    # Error API.
    PathrsError,
    _is_pathrs_err,
    INTERNAL_ERROR,
    # CFFI helpers.
    _cstr,
    _pystr,
    _cbuffer,
)
from ._libpathrs_cffi import lib as libpathrs_so

if typing.TYPE_CHECKING:
    # mypy apparently cannot handle the "ffi: cffi.api.FFI" definition in
    # _libpathrs_cffi/__init__.pyi so we need to explicitly reference the type
    # from cffi here.
    import cffi

    ffi = cffi.FFI()
    CString: TypeAlias = cffi.FFI.CData
    CBuffer: TypeAlias = cffi.FFI.CData
else:
    from ._libpathrs_cffi import ffi

    CString: TypeAlias = ffi.CData
    CBuffer: TypeAlias = ffi.CData

__all__ = [
    # Core api.
    "Root",
    "Handle",
    "library_version",
    # Error api (re-export).
    "PathrsError",
]


class VersionInfo(metaclass=SingletonClass):
    """Stores information about the runtime version of libpathrs.so."""

    _VERSION_INFO_TYPE = "pathrs_version_info_t *"

    version_string: str

    def __init__(self):  # type: ignore[no-untyped-def] # TODO: mypy appears to struggle with metaclasses.
        info = cast("libpathrs_so.VersionInfo", ffi.new(self._VERSION_INFO_TYPE))
        info_size = ffi.sizeof(cast("Any", info))

        size = libpathrs_so.pathrs_version(info, info_size)
        if _is_pathrs_err(size):
            raise PathrsError._fetch(size) or INTERNAL_ERROR
        if size > 0:
            # TODO: We should also output a warning if the size of
            # pathrs_version_info_t is bigger than the number of fields in the
            # python-native VersionInfo. Otherwise a rebuild will mask that
            # Python callers cannot see any new fields.
            warnings.warn(
                f"pathrs_version returned extra data we could not parse (need {size} bytes but buffer was only {info_size})"
            )

        self.version_string = _pystr(info.version_string)


def library_version() -> VersionInfo:
    """
    Returns information about the runtime version of libpathrs.so.
    This is just shorthand for VersionInfo().
    """
    return VersionInfo()  # type: ignore[no-untyped-call] # TODO: mypy appears to struggle with metaclasses.


class Handle(WrappedFd):
    "A handle to a filesystem object, usually resolved using Root.resolve()."

    def reopen(self, mode: str = "r", /, *, extra_flags: int = 0) -> IO[Any]:
        """
        Upgrade a Handle to a os.fdopen() file handle.

        mode is a Python mode string, and extra_flags can be used to indicate
        extra O_* flags you wish to pass to the reopen operation.

        The returned file handle is independent to the original Handle, and you
        can freely call Handle.reopen() on the same Handle multiple times.
        """
        flags = _convert_mode(mode) | extra_flags
        with self.reopen_raw(flags) as file:
            return file.fdopen(mode)

    def reopen_raw(self, flags: int, /) -> WrappedFd:
        """
        Upgrade a Handle to a WrappedFd file handle.

        flags is the set of O_* flags you wish to pass to the open operation.

        The returned file handle is independent to the original Handle, and you
        can freely call Handle.reopen() on the same Handle multiple times.
        """
        fd = libpathrs_so.pathrs_reopen(self.fileno(), flags)
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return WrappedFd(fd)


class Root(WrappedFd):
    """
    A handle to a filesystem root, which filesystem operations are all done
    relative to.
    """

    def __init__(self, file_or_path: Union[FileLike, str], /):
        """
        Create a handle from a file-like object or a path to a directory.

        Note that creating a Root in an attacker-controlled directory can allow
        for an attacker to trick you into allowing breakouts. If file_or_path
        is a path string, be aware there are no protections against rename race
        attacks when opening the Root directory handle itself.
        """
        if isinstance(file_or_path, str):
            path = _cstr(file_or_path)
            fd = libpathrs_so.pathrs_open_root(path)
            if _is_pathrs_err(fd):
                raise PathrsError._fetch(fd) or INTERNAL_ERROR
            file: FileLike = fd
        else:
            file = file_or_path

        # XXX: Is this necessary?
        super().__init__(file)

    def resolve(self, path: str, /, *, follow_trailing: bool = True) -> Handle:
        """
        Resolve the given path inside the Root and return a Handle.

        follow_trailing indicates what resolve should do if the final component
        of the path is a symlink. The default is to continue resolving it, if
        follow_trailing=False then a handle to the symlink itself is returned.
        This has some limited uses, but most users should use the default.

        A pathrs.Error is raised if the path doesn't exist.
        """
        if follow_trailing:
            fd = libpathrs_so.pathrs_inroot_resolve(self.fileno(), _cstr(path))
        else:
            fd = libpathrs_so.pathrs_inroot_resolve_nofollow(self.fileno(), _cstr(path))
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return Handle(fd)

    def open(
        self,
        path: str,
        mode: str = "r",
        /,
        *,
        follow_trailing: bool = True,
        extra_flags: int = 0,
    ) -> IO[Any]:
        """
        Resolve and open a path inside the Root and return an os.fdopen() file
        handle.

        This is effectively shorthand for Root.resolve(path).reopen(...), but
        for the openat2-based resolver this is slightly more efficient if you
        just want to open a file and don't need to do multiple reopens of the
        Handle.

        mode is a Python mode string, and extra_flags can be used to indicate
        extra O_* flags you wish to pass to the open operation.

        follow_trailing has the same behaviour as Root.resolve(), and is
        equivalent to passing os.O_NOFOLLOW to extra_flags.
        """
        flags = _convert_mode(mode) | extra_flags
        if not follow_trailing:
            flags |= os.O_NOFOLLOW
        with self.open_raw(path, flags) as file:
            return file.fdopen(mode)

    def open_raw(self, path: str, flags: int, /) -> WrappedFd:
        """
        Resolve and open a path inside the Root and return a WrappedFd file
        handle.

        This is effectively shorthand for Root.resolve(path).reopen_raw(flags),
        but for the openat2-based resolver this is slightly more efficient if
        you just want to open a file and don't need to do multiple reopens of
        the Handle.

        If flags contains os.O_NOFOLLOW, then the resolution is done as if you
        passed follow_trailing=False to Root.resolve().
        """
        fd = libpathrs_so.pathrs_inroot_open(self.fileno(), _cstr(path), flags)
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return WrappedFd(fd)

    def readlink(self, path: str, /) -> str:
        """
        Fetch the target of a symlink at the given path in the Root.

        A pathrs.Error is raised if the path is not a symlink or doesn't exist.
        """
        cpath = _cstr(path)
        linkbuf_size = 128
        while True:
            linkbuf = _cbuffer(linkbuf_size)
            n = libpathrs_so.pathrs_inroot_readlink(
                self.fileno(), cpath, linkbuf, linkbuf_size
            )
            if _is_pathrs_err(n):
                raise PathrsError._fetch(n) or INTERNAL_ERROR
            elif n <= linkbuf_size:
                buf = typing.cast(bytes, ffi.buffer(linkbuf, linkbuf_size)[:n])
                return buf.decode("latin1")
            else:
                # The contents were truncated. Unlike readlinkat, pathrs returns
                # the size of the link when it checked. So use the returned size
                # as a basis for the reallocated size (but in order to avoid a DoS
                # where a magic-link is growing by a single byte each iteration,
                # make sure we are a fair bit larger).
                linkbuf_size += n

    def creat(
        self, path: str, mode: str = "r", filemode: int = 0o644, /, extra_flags: int = 0
    ) -> IO[Any]:
        """
        Atomically create-and-open a new file at the given path in the Root,
        a-la O_CREAT.

        This method returns an os.fdopen() file handle.

        filemode is the Unix DAC mode you wish the new file to be created with.
        This mode might not be the actual mode of the created file due to a
        variety of external factors (umask, setgid bits, POSIX ACLs).

        mode is a Python mode string, and extra_flags can be used to indicate
        extra O_* flags you wish to pass to the reopen operation. If you wish
        to ensure the new file was created *by you* then you may wish to add
        O_EXCL to extra_flags.
        """
        flags = _convert_mode(mode) | extra_flags
        fd = libpathrs_so.pathrs_inroot_creat(
            self.fileno(), _cstr(path), flags, filemode
        )
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return os.fdopen(fd, mode)

    def creat_raw(self, path: str, flags: int, filemode: int = 0o644, /) -> WrappedFd:
        """
        Atomically create-and-open a new file at the given path in the Root,
        a-la O_CREAT.

        This method returns a WrappedFd handle.

        filemode is the Unix DAC mode you wish the new file to be created with.
        This mode might not be the actual mode of the created file due to a
        variety of external factors (umask, setgid bits, POSIX ACLs).

        flags is the set of O_* flags you wish to pass to the open operation. If
        you do not intend to open a symlink, you should pass O_NOFOLLOW to flags to
        let libpathrs know that it can be more strict when opening the path.
        """
        fd = libpathrs_so.pathrs_inroot_creat(
            self.fileno(), _cstr(path), flags, filemode
        )
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return WrappedFd(fd)

    def rename(self, src: str, dst: str, flags: int = 0, /) -> None:
        """
        Rename a path from src to dst within the Root.

        flags can be any renameat2(2) flags you wish to use, which can change
        the behaviour of this method substantially. For instance,
        RENAME_EXCHANGE will turn this into an atomic swap operation.
        """
        # TODO: Should we have a separate Root.swap() operation?
        err = libpathrs_so.pathrs_inroot_rename(
            self.fileno(), _cstr(src), self.fileno(), _cstr(dst), flags
        )
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR

    def rmdir(self, path: str, /) -> None:
        """
        Remove an empty directory at the given path within the Root.

        To remove non-empty directories recursively, you can use
        Root.remove_all().
        """
        err = libpathrs_so.pathrs_inroot_rmdir(self.fileno(), _cstr(path))
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR

    def unlink(self, path: str, /) -> None:
        """
        Remove a non-directory inode at the given path within the Root.

        To remove empty directories, you can use Root.remove_all(). To remove
        files and non-empty directories recursively, you can use
        Root.remove_all().
        """
        err = libpathrs_so.pathrs_inroot_unlink(self.fileno(), _cstr(path))
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR

    def remove_all(self, path: str, /) -> None:
        """
        Remove the file or directory (empty or non-empty) at the given path
        within the Root.
        """
        err = libpathrs_so.pathrs_inroot_remove_all(self.fileno(), _cstr(path))
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR

    def mkdir(self, path: str, mode: int, /) -> None:
        """
        Create a directory at the given path within the Root.

        mode is the Unix DAC mode you wish the new directory to be created
        with. This mode might not be the actual mode of the created file due to
        a variety of external factors (umask, setgid bits, POSIX ACLs).

        A pathrs.Error will be raised if the parent directory doesn't exist, or
        the path already exists. To create a directory and all of its parent
        directories (or just reuse an existing directory) you can use
        Root.mkdir_all().
        """
        err = libpathrs_so.pathrs_inroot_mkdir(self.fileno(), _cstr(path), mode)
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR

    def mkdir_all(self, path: str, mode: int, /) -> Handle:
        """
        Recursively create a directory and all of its parents at the given path
        within the Root (or reuse an existing directory if the path already
        exists).

        This method returns a Handle to the created directory.

        mode is the Unix DAC mode you wish any new directories to be created
        with. This mode might not be the actual mode of the created file due to
        a variety of external factors (umask, setgid bits, POSIX ACLs). If the
        full path already exists, this mode is ignored and the existing
        directory mode is kept.
        """
        fd = libpathrs_so.pathrs_inroot_mkdir_all(self.fileno(), _cstr(path), mode)
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return Handle(fd)

    def mknod(self, path: str, mode: int, device: int = 0, /) -> None:
        """
        Create a new inode at the given path within the Root.

        mode both indicates the file type (it must contain a valid bit from
        S_IFMT to indicate what kind of file to create) and what the mode of
        the newly created file should have. This mode might not be the actual
        mode of the created file due to a variety of external factors (umask,
        setgid bits, POSIX ACLs).

        dev is the the (major, minor) device number used for the new inode if
        the mode contains S_IFCHR or S_IFBLK. You can construct the device
        number from a (major, minor) using os.makedev().

        A pathrs.Error is raised if the path already exists.
        """
        err = libpathrs_so.pathrs_inroot_mknod(self.fileno(), _cstr(path), mode, device)
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR

    def hardlink(self, target: str, linkname: str, /) -> None:
        """
        Create a hardlink between two paths inside the Root.

        linkname is the path to the *new* hardlink, and target is a path to
        the *existing* file.

        A pathrs.Error is raised if the path for the new hardlink already
        exists.
        """
        err = libpathrs_so.pathrs_inroot_hardlink(
            self.fileno(), _cstr(target), self.fileno(), _cstr(linkname), 0
        )
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR

    def symlink(self, target: str, linkname: str, /) -> None:
        """
        Create a symlink at the given path in the Root.

        linkname is the path to the *new* symlink, and target is what the
        symlink will point to. Note that symlink contents are not verified on
        Linux, so there are no restrictions on what target you put.

        A pathrs.Error is raised if the path for the new symlink already
        exists.
        """
        err = libpathrs_so.pathrs_inroot_symlink(
            _cstr(target), self.fileno(), _cstr(linkname)
        )
        if _is_pathrs_err(err):
            raise PathrsError._fetch(err) or INTERNAL_ERROR



#!/usr/bin/python3
# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import setuptools

from typing import Any, Dict


# This is only needed for backwards compatibility with older versions.
def parse_pyproject() -> Dict[str, Any]:
    try:
        import tomllib

        openmode = "rb"
    except ImportError:
        # TODO: Remove this once we only support Python >= 3.11.
        import toml as tomllib  # type: ignore # The types are not compatible.

        openmode = "r"

    with open("pyproject.toml", openmode) as f:
        return tomllib.load(f)


pyproject = parse_pyproject()

setuptools.setup(
    # For backwards-compatibility with pre-pyproject setuptools.
    name=pyproject["project"]["name"],
    version=pyproject["project"]["version"],
    install_requires=pyproject["project"]["dependencies"],
    # Configure cffi building.
    ext_package="pathrs",
    platforms=["Linux"],
    cffi_modules=["pathrs/pathrs_build.py:ffibuilder"],
)



# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import cffi

ffi: cffi.FFI



#!/usr/bin/python3
# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing
from typing import Any, IO, cast

# TODO: Remove this once we only support Python >= 3.11.
from typing_extensions import Self, TypeAlias

from ._internal import (
    # File type helpers.
    WrappedFd,
    _convert_mode,
    # Error API.
    PathrsError,
    _is_pathrs_err,
    INTERNAL_ERROR,
    # CFFI helpers.
    _cstr,
    _cbuffer,
)
from ._libpathrs_cffi import lib as libpathrs_so

if typing.TYPE_CHECKING:
    # mypy apparently cannot handle the "ffi: cffi.api.FFI" definition in
    # _libpathrs_cffi/__init__.pyi so we need to explicitly reference the type
    # from cffi here.
    import cffi

    ffi = cffi.FFI()
    CString: TypeAlias = cffi.FFI.CData
    CBuffer: TypeAlias = cffi.FFI.CData
else:
    from ._libpathrs_cffi import ffi

    CString: TypeAlias = ffi.CData
    CBuffer: TypeAlias = ffi.CData

__all__ = [
    "PROC_ROOT",
    "PROC_SELF",
    "PROC_THREAD_SELF",
    "PROC_PID",
    "ProcfsHandle",
    # Shorthand for ProcfsHandle.cached().<foo>.
    "open",
    "open_raw",
    "readlink",
]

# TODO: Switch to "type ..." syntax once we switch to Python >= 3.12...?
ProcfsBase: TypeAlias = int

#: Resolve proc_* operations relative to the /proc root. Note that this mode
#: may be more expensive because we have to take steps to try to avoid leaking
#: unmasked procfs handles, so you should use PROC_SELF if you can.
PROC_ROOT: ProcfsBase = libpathrs_so.PATHRS_PROC_ROOT

#: Resolve proc_* operations relative to /proc/self. For most programs, this is
#: the standard choice.
PROC_SELF: ProcfsBase = libpathrs_so.PATHRS_PROC_SELF

#: Resolve proc_* operations relative to /proc/thread-self. In multi-threaded
#: programs where one thread has a different CLONE_FS, it is possible for
#: /proc/self to point the wrong thread and so /proc/thread-self may be
#: necessary.
PROC_THREAD_SELF: ProcfsBase = libpathrs_so.PATHRS_PROC_THREAD_SELF


def PROC_PID(pid: int) -> ProcfsBase:
    """
    Resolve proc_* operations relative to /proc/<pid>. Be aware that due to PID
    recycling, using this is generally not safe except in certain
    circumstances. Namely:

     * PID 1 (the init process), as that PID cannot ever get recycled.
     * Your current PID (though you should just use PROC_SELF).
     * PIDs of child processes (as long as you are sure that no other part of
       your program incorrectly catches or ignores SIGCHLD, and that you do it
       *before* you call wait(2)or any equivalent method that could reap
       zombies).
    """
    if pid & libpathrs_so.__PATHRS_PROC_TYPE_MASK:
        raise ValueError(f"invalid PROC_PID value {pid}")
    return libpathrs_so.__PATHRS_PROC_TYPE_PID | pid


class ProcfsHandle(WrappedFd):
    """ """

    _PROCFS_OPEN_HOW_TYPE = "pathrs_procfs_open_how *"

    @classmethod
    def cached(cls) -> Self:
        """
        Returns a cached version of ProcfsHandle that will always remain valid
        and cannot be closed. This is the recommended usage of ProcfsHandle.
        """
        return cls.from_raw_fd(libpathrs_so.PATHRS_PROC_DEFAULT_ROOTFD)

    @classmethod
    def new(cls, /, *, unmasked: bool = False) -> Self:
        """
        Create a new procfs handle with the requested configuration settings.

        Note that the requested configuration might be eligible for caching, in
        which case the ProcfsHandle.fileno() will contain a special sentinel
        value that cannot be used like a regular file descriptor.
        """

        # TODO: Is there a way to have ProcfsOpenHow actually subclass CData so
        # that we don't need to do any of these ugly casts?
        how = cast("libpathrs_so.ProcfsOpenHow", ffi.new(cls._PROCFS_OPEN_HOW_TYPE))
        how_size = ffi.sizeof(cast("Any", how))

        if unmasked:
            how.flags = libpathrs_so.PATHRS_PROCFS_NEW_UNMASKED

        fd = libpathrs_so.pathrs_procfs_open(how, how_size)
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return cls.from_raw_fd(fd)

    def open_raw(self, base: ProcfsBase, path: str, flags: int, /) -> WrappedFd:
        """
        Open a procfs file using Unix open flags.

        This function returns a WrappedFd file handle.

        base indicates what the path should be relative to. Valid values
        include PROC_{ROOT,SELF,THREAD_SELF}.

        path is a relative path to base indicating which procfs file you wish
        to open.

        flags is the set of O_* flags you wish to pass to the open operation.
        If you do not intend to open a symlink, you should pass O_NOFOLLOW to
        flags to let libpathrs know that it can be more strict when opening the
        path.
        """
        # TODO: Should we default to O_NOFOLLOW or put a separate argument for it?
        fd = libpathrs_so.pathrs_proc_openat(self.fileno(), base, _cstr(path), flags)
        if _is_pathrs_err(fd):
            raise PathrsError._fetch(fd) or INTERNAL_ERROR
        return WrappedFd(fd)

    def open(
        self, base: ProcfsBase, path: str, mode: str = "r", /, *, extra_flags: int = 0
    ) -> IO[Any]:
        """
        Open a procfs file using Pythonic mode strings.

        This function returns an os.fdopen() file handle.

        base indicates what the path should be relative to. Valid values
        include PROC_{ROOT,SELF,THREAD_SELF}.

        path is a relative path to base indicating which procfs file you wish
        to open.

        mode is a Python mode string, and extra_flags can be used to indicate
        extra O_* flags you wish to pass to the open operation. If you do not
        intend to open a symlink, you should pass O_NOFOLLOW to extra_flags to
        let libpathrs know that it can be more strict when opening the path.
        """
        flags = _convert_mode(mode) | extra_flags
        with self.open_raw(base, path, flags) as file:
            return file.fdopen(mode)

    def readlink(self, base: ProcfsBase, path: str, /) -> str:
        """
        Fetch the target of a procfs symlink.

        Note that some procfs symlinks are "magic-links" where the returned
        string from readlink() is not how they are actually resolved.

        base indicates what the path should be relative to. Valid values
        include PROC_{ROOT,SELF,THREAD_SELF}.

        path is a relative path to base indicating which procfs file you wish
        to open.
        """
        # TODO: See if we can merge this with Root.readlink.
        cpath = _cstr(path)
        linkbuf_size = 128
        while True:
            linkbuf = _cbuffer(linkbuf_size)
            n = libpathrs_so.pathrs_proc_readlinkat(
                self.fileno(), base, cpath, linkbuf, linkbuf_size
            )
            if _is_pathrs_err(n):
                raise PathrsError._fetch(n) or INTERNAL_ERROR
            elif n <= linkbuf_size:
                buf = typing.cast(bytes, ffi.buffer(linkbuf, linkbuf_size)[:n])
                return buf.decode("latin1")
            else:
                # The contents were truncated. Unlike readlinkat, pathrs
                # returns the size of the link when it checked. So use the
                # returned size as a basis for the reallocated size (but in
                # order to avoid a DoS where a magic-link is growing by a
                # single byte each iteration, make sure we are a fair bit
                # larger).
                linkbuf_size += n


#: Open a procfs file (with unix open flags).
#: Shorthand for ProcfsHandle.cached().open(...).
open = ProcfsHandle.cached().open

#: Open a procfs file (with Pythonic mode strings).
#: Shorthand for ProcfsHandle.cached().open_raw(...).
open_raw = ProcfsHandle.cached().open_raw

#: Fetch the target of a procfs symlink.
#: Shorthand for ProcfsHandle.cached().readlink(...).
readlink = ProcfsHandle.cached().readlink



# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import type_check_only, Union

# TODO: Remove this once we only support Python >= 3.10.
from typing_extensions import TypeAlias, Literal

from .._pathrs import CBuffer, CString
from ..procfs import ProcfsBase

RawFd: TypeAlias = int

# pathrs_errorinfo_t *
@type_check_only
class CError:
    saved_errno: int
    description: CString

ErrorId: TypeAlias = int
__PATHRS_MAX_ERR_VALUE: ErrorId

# TODO: We actually return Union[CError, cffi.FFI.NULL] but we can't express
#       this using the typing stubs for CFFI...
def pathrs_errorinfo(err_id: Union[ErrorId, int]) -> CError: ...
def pathrs_errorinfo_free(err: CError) -> None: ...

# pathrs_version_info_t *
@type_check_only
class VersionInfo:
    version_string: CString

def pathrs_version(info: VersionInfo, size: int) -> Union[ErrorId, int]: ...

# uint64_t
ProcfsOpenFlags: TypeAlias = int
PATHRS_PROCFS_NEW_UNMASKED: ProcfsOpenFlags

# pathrs_procfs_open_how *
@type_check_only
class ProcfsOpenHow:
    flags: ProcfsOpenFlags

PATHRS_PROC_ROOT: ProcfsBase
PATHRS_PROC_SELF: ProcfsBase
PATHRS_PROC_THREAD_SELF: ProcfsBase

__PATHRS_PROC_TYPE_MASK: ProcfsBase
__PATHRS_PROC_TYPE_PID: ProcfsBase

PATHRS_PROC_DEFAULT_ROOTFD: RawFd

# procfs API
def pathrs_procfs_open(how: ProcfsOpenHow, size: int) -> Union[RawFd, ErrorId]: ...
def pathrs_proc_open(
    base: ProcfsBase, path: CString, flags: int
) -> Union[RawFd, ErrorId]: ...
def pathrs_proc_openat(
    proc_root_fd: RawFd, base: ProcfsBase, path: CString, flags: int
) -> Union[RawFd, ErrorId]: ...
def pathrs_proc_readlink(
    base: ProcfsBase, path: CString, linkbuf: CBuffer, linkbuf_size: int
) -> Union[int, ErrorId]: ...
def pathrs_proc_readlinkat(
    proc_root_fd: RawFd,
    base: ProcfsBase,
    path: CString,
    linkbuf: CBuffer,
    linkbuf_size: int,
) -> Union[int, ErrorId]: ...

# core API
def pathrs_open_root(path: CString) -> Union[RawFd, ErrorId]: ...
def pathrs_reopen(fd: RawFd, flags: int) -> Union[RawFd, ErrorId]: ...
def pathrs_inroot_resolve(rootfd: RawFd, path: CString) -> Union[RawFd, ErrorId]: ...
def pathrs_inroot_resolve_nofollow(
    rootfd: RawFd, path: CString
) -> Union[RawFd, ErrorId]: ...
def pathrs_inroot_open(
    rootfd: RawFd, path: CString, flags: int
) -> Union[RawFd, ErrorId]: ...
def pathrs_inroot_creat(
    rootfd: RawFd, path: CString, flags: int, filemode: int
) -> Union[RawFd, ErrorId]: ...
def pathrs_inroot_rename(
    old_rootfd: RawFd,
    old_path: CString,
    new_rootfd: RawFd,
    new_path: CString,
    flags: int,
) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_rmdir(rootfd: RawFd, path: CString) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_unlink(
    rootfd: RawFd, path: CString
) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_remove_all(rootfd: RawFd, path: CString) -> Union[RawFd, ErrorId]: ...
def pathrs_inroot_mkdir(
    rootfd: RawFd, path: CString, mode: int
) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_mkdir_all(
    rootfd: RawFd, path: CString, mode: int
) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_mknod(
    rootfd: RawFd, path: CString, mode: int, dev: int
) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_hardlink(
    old_rootfd: RawFd,
    old_path: CString,
    new_rootfd: RawFd,
    new_path: CString,
    flags: int,
) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_symlink(
    target: CString,
    rootfd: RawFd,
    linkpath: CString,
) -> Union[Literal[0], ErrorId]: ...
def pathrs_inroot_readlink(
    rootfd: RawFd, path: CString, linkbuf: CBuffer, linkbuf_size: int
) -> Union[int, ErrorId]: ...



#!/usr/bin/python3
# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import importlib
import importlib.metadata

from . import _pathrs
from ._pathrs import *  # noqa: F403 # We just re-export everything.

# In order get pydoc to include the documentation for the re-exported code from
# _pathrs, we need to include all of the members in __all__. Rather than
# duplicating the member list here explicitly, just re-export __all__.
__all__ = []
__all__ += _pathrs.__all__  # pyright doesn't support "=" here.

try:
    # In order to avoid drift between this version and the dist-info/ version
    # information, just fill __version__ with the dist-info/ information.
    __version__ = importlib.metadata.version("pathrs")
except importlib.metadata.PackageNotFoundError:
    # We're being run from a local directory without an installed version of
    # pathrs, so just fill in a dummy version.
    __version__ = "<unknown>"



#!/usr/bin/python3
# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

# This builds the _pathrs module (only needs to be done during the initial
# build of libpathrs, and can be redistributed alongside the pathrs.py wrapping
# library). It's much better than the ABI-mode of CFFI.

import re
import os
import sys

from typing import Any, Optional
from collections.abc import Iterable

import cffi


def load_hdr(ffi: cffi.FFI, hdr_path: str) -> None:
    with open(hdr_path) as f:
        hdr = f.read()

    # We need to first filter out all the bits of <pathrs.h> that are not
    # supported by cffi. Ideally this wouldn't be necessary, but the macro
    # support in cffi is very limited, and we make use of basic features that
    # are unsupported by cffi.

    # Drop all non-#define lines (directives are not supported).
    hdr = re.sub(r"^#\s*(?!define\b).*$", "", hdr, flags=re.MULTILINE)
    # Drop all:
    #  * "#define FOO(n) ..." lines (function-style macros are not supported).
    #  * Empty-value "#define FOO" lines (empty macros are not supported)
    # TODO: We probably should support multi-line macros.
    hdr = re.sub(r"^#\s*define\b\s*\w*(\(.*|)$", "", hdr, flags=re.MULTILINE)

    # Replace each struct-like body that has __CBINDGEN_ALIGNED before it,
    # remove the __CBINDGEN_ALIGNED and add "...;" as the last field in the
    # struct. This is how you tell cffi to get the proper alignment from the
    # compiler (__attribute__((aligned(n))) is not supported by cdef).
    hdr = re.sub(
        r"__CBINDGEN_ALIGNED\(\d+\)([^{;]*){([^}]+)}",
        r"\1 {\2 ...;}",
        hdr,
        flags=re.MULTILINE,
    )

    # Load the header.
    ffi.cdef(hdr)


def create_ffibuilder(**kwargs: Any) -> cffi.FFI:
    ffibuilder = cffi.FFI()
    ffibuilder.cdef("typedef uint32_t dev_t;")

    # We need to use cdef to tell cffi what functions we need to FFI to. But we
    # don't need the structs (I hope).
    for include_dir in kwargs.get("include_dirs", []):
        pathrs_hdr = os.path.join(include_dir, "pathrs.h")
        if os.path.exists(pathrs_hdr):
            load_hdr(ffibuilder, pathrs_hdr)

    # Add a source and link to libpathrs.
    ffibuilder.set_source(
        "_libpathrs_cffi", "#include <pathrs.h>", libraries=["pathrs"], **kwargs
    )

    return ffibuilder


def find_rootdir() -> str:
    # Figure out where the libpathrs source dir is.
    root_dir = None
    candidate = os.path.dirname(sys.path[0] or os.getcwd())
    while candidate != "/":
        try:
            # Look for a Cargo.toml which says it's pathrs.
            candidate_toml = os.path.join(candidate, "Cargo.toml")
            with open(candidate_toml, "r") as f:
                content = f.read()
            if re.findall(r'^name = "pathrs"$', content, re.MULTILINE):
                root_dir = candidate
                break
        except FileNotFoundError:
            pass
        candidate = os.path.dirname(candidate)

    if not root_dir:
        raise FileNotFoundError("Could not find pathrs source-dir root.")

    return root_dir


def srcdir_ffibuilder(root_dir: Optional[str] = None) -> cffi.FFI:
    """
    Build the CFFI bindings using the provided root_dir as the root of a
    pathrs source tree which has compiled cdylibs ready in target/*.
    """

    if root_dir is None:
        root_dir = find_rootdir()

    # Figure out which libs are usable.
    library_dirs: Iterable[str] = (
        os.path.join(root_dir, "target/%s/libpathrs.so" % (mode,))
        for mode in ("debug", "release")
    )
    library_dirs = (so_path for so_path in library_dirs if os.path.exists(so_path))
    library_dirs = sorted(library_dirs, key=lambda path: -os.path.getmtime(path))
    library_dirs = [os.path.dirname(path) for path in library_dirs]

    # Compile the libpathrs module.
    return create_ffibuilder(
        include_dirs=[os.path.join(root_dir, "include")],
        library_dirs=library_dirs,
    )


def system_ffibuilder() -> cffi.FFI:
    """
    Build the CFFI bindings using the installed libpathrs system libraries.
    """

    return create_ffibuilder(
        include_dirs=[
            "/usr/include",
            "/usr/local/include",
        ]
    )


if __name__ == "__main__":
    try:
        # Search for the compiled libraries to link to from our libpathrs
        # source if running outside of setuptools as a regular program.
        ffibuilder = srcdir_ffibuilder(root_dir=find_rootdir())
    except FileNotFoundError:
        # If we couldn't find a valid library in the source dir, just fallback
        # to using the system libraries.
        ffibuilder = system_ffibuilder()
    ffibuilder.compile(verbose=True)
elif os.environ.get("PATHRS_SRC_ROOT", "") != "":
    # If we're running in setup tools, we can't easily find the source dir.
    # However, distributions can set PATHRS_SRC_ROOT to the path of the
    # libpathrs source directory to make it easier to build the python modules
    # in the same %build script as the main library.
    ffibuilder = srcdir_ffibuilder(root_dir=os.environ.get("PATHRS_SRC_ROOT"))
else:
    # Use the system libraries if running inside standard setuptools.
    ffibuilder = system_ffibuilder()



#!/usr/bin/env python3
# SPDX-License-Identifier: MPL-2.0
#
# libpathrs: safe path resolution on Linux
# Copyright (C) 2019-2025 SUSE LLC
# Copyright (C) 2026 Aleksa Sarai <cyphar@cyphar.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import os
import sys
import stat
import argparse
from typing import Optional, Self, Sequence, Protocol, Tuple

sys.path.append(os.path.dirname(__file__) + "/../../../contrib/bindings/python")
import pathrs
from pathrs import procfs
from pathrs.procfs import ProcfsHandle, ProcfsBase


def version(args: argparse.Namespace):
    version = pathrs.library_version()
    print("VERSION", version.version_string)


class FilenoFile(Protocol):
    def fileno(self) -> int: ...


def fdpath(fd: FilenoFile) -> str:
    return ProcfsHandle.cached().readlink(procfs.PROC_THREAD_SELF, f"fd/{fd.fileno()}")


def root_resolve(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath
    follow: bool = args.follow
    reopen: Optional[int] = args.reopen

    with root.resolve(subpath, follow_trailing=follow) as handle:
        print("HANDLE-PATH", fdpath(handle))
        if reopen is not None:
            with handle.reopen_raw(reopen) as file:
                print("FILE-PATH", fdpath(file))


def root_open(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath
    oflags: int = args.oflags
    follow: bool = args.follow

    if not follow:
        oflags |= os.O_NOFOLLOW

    with root.open_raw(subpath, oflags) as file:
        print("FILE-PATH", fdpath(file))


def root_mkfile(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath
    oflags: int = args.oflags
    mode: int = args.mode

    with root.creat_raw(subpath, oflags, mode) as file:
        print("FILE-PATH", fdpath(file))


def root_mkdir(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath
    mode: int = args.mode

    root.mkdir(subpath, mode)


def root_mkdir_all(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath
    mode: int = args.mode

    with root.mkdir_all(subpath, mode) as handle:
        print("HANDLE-PATH", fdpath(handle))


def root_mknod(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath
    mode: int = args.mode
    inode_type: int
    inode_dev: int
    (inode_type, inode_dev) = args.type

    mode |= inode_type
    root.mknod(subpath, mode, inode_dev)


def root_hardlink(args: argparse.Namespace):
    root: pathrs.Root = args.root
    target: str = args.target
    linkname: str = args.linkname

    root.hardlink(target, linkname)


def root_symlink(args: argparse.Namespace):
    root: pathrs.Root = args.root
    target: str = args.target
    linkname: str = args.linkname

    root.symlink(target, linkname)


def root_readlink(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath

    target = root.readlink(subpath)
    print("LINK-TARGET", target)


def root_unlink(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath

    root.unlink(subpath)


def root_rmdir(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath

    root.rmdir(subpath)


def root_rmdir_all(args: argparse.Namespace):
    root: pathrs.Root = args.root
    subpath: str = args.subpath

    root.remove_all(subpath)


def root_rename(args: argparse.Namespace):
    root: pathrs.Root = args.root
    source: str = args.source
    destination: str = args.destination

    clobber: bool = args.clobber
    exchange: bool = args.exchange
    whiteout: bool = args.whiteout

    RENAME_NOREPLACE: int = 1 << 0
    RENAME_EXCHANGE: int = 1 << 1
    RENAME_WHITEOUT: int = 1 << 2

    rename_args: int = 0
    if not clobber:
        rename_args |= RENAME_NOREPLACE
    if exchange:
        rename_args |= RENAME_EXCHANGE
    if whiteout:
        rename_args |= RENAME_WHITEOUT

    root.rename(source, destination, rename_args)


def procfs_open(args: argparse.Namespace):
    unmasked: bool = args.unmasked
    proc = ProcfsHandle.new(unmasked=unmasked)

    base: ProcfsBase = args.procfs_base
    subpath: str = args.subpath
    oflags: int = args.oflags
    follow: bool = args.follow

    if not follow:
        oflags |= os.O_NOFOLLOW

    with proc.open_raw(base, subpath, oflags) as file:
        print("FILE-PATH", fdpath(file))


def procfs_readlink(args: argparse.Namespace):
    unmasked: bool = args.unmasked
    proc = ProcfsHandle.new(unmasked=unmasked)

    base: ProcfsBase = args.procfs_base
    subpath: str = args.subpath

    target = proc.readlink(base, subpath)
    print("LINK-TARGET", target)


def parse_args(
    args: tuple[str, ...],
) -> Tuple[argparse.ArgumentParser, argparse.Namespace]:
    parser = argparse.ArgumentParser(prog="pathrs-cmd")
    parser.set_defaults(func=None)
    top_subparser = parser.add_subparsers()

    def add_mode_flag(
        parser: argparse.ArgumentParser,
        name: str,
        default: int = 0o644,
        required: bool = False,
        help: Optional[str] = None,
    ) -> None:
        parser.add_argument(
            f"--{name}",
            type=lambda mode: int(mode, 8),
            default=default,
            required=required,
            help=help,
        )

    # TODO: Should this be a class?
    def parse_oflags(flags: str) -> int:
        VALID_FLAGS = {
            # Access modes (including O_PATH).
            "O_RDWR": os.O_RDWR,
            "O_RDONLY": os.O_RDONLY,
            "O_WRONLY": os.O_WRONLY,
            "O_PATH": os.O_PATH,
            # Fd flags.
            "O_CLOEXEC": os.O_CLOEXEC,
            # Control lookups.
            "O_NOFOLLOW": os.O_NOFOLLOW,
            "O_DIRECTORY": os.O_DIRECTORY,
            "O_NOCTTY": os.O_NOCTTY,
            # NOTE: This flag contains O_DIRECTORY!
            "O_TMPFILE": os.O_TMPFILE,
            # File creation.
            "O_CREAT": os.O_CREAT,
            "O_EXCL": os.O_EXCL,
            "O_TRUNC": os.O_TRUNC,
            "O_APPEND": os.O_APPEND,
            # Sync.
            "O_SYNC": os.O_SYNC,
            "O_ASYNC": os.O_ASYNC,
            "O_DSYNC": os.O_DSYNC,
            "O_FSYNC": os.O_FSYNC,
            "O_RSYNC": os.O_RSYNC,
            "O_DIRECT": os.O_DIRECT,
            "O_NDELAY": os.O_NDELAY,
            "O_NOATIME": os.O_NOATIME,
            "O_NONBLOCK": os.O_NONBLOCK,
        }

        all_flags = 0
        for flag in flags.replace(",", "|").split("|"):
            flag = flag.upper()
            if not flag.startswith("O_"):
                flag = "O_" + flag
            try:
                all_flags |= VALID_FLAGS[flag]
            except KeyError:
                raise ValueError(f"flag {flag:r} is not a valid O_* flag")
        return all_flags

    def add_o_flag(
        parser: argparse.ArgumentParser,
        name: str,
        default: Optional[int] = os.O_RDONLY,
        required: bool = False,
        help: Optional[str] = None,
    ) -> None:
        parser.add_argument(
            f"--{name}",
            metavar="O_*",
            type=parse_oflags,
            default=default,
            required=required,
            help=f"{help or name} (comma- or |-separated)",
        )

    version_parser = top_subparser.add_parser("version", help="version information")
    version_parser.set_defaults(func=version)
    del version_parser

    # root --root <root> ... commands
    root_parser = top_subparser.add_parser("root", help="Root::* operations")
    root_parser.add_argument(
        "--root", type=pathrs.Root, required=True, help="root path"
    )
    root_subparser = root_parser.add_subparsers()

    # root resolve [--reopen <flags>] [--[no-]follow] subpath
    root_resolve_parser = root_subparser.add_parser(
        "resolve", help="resolve a path inside the root"
    )
    root_resolve_parser.add_argument(
        "--follow",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="follow trailing symlinks",
    )
    add_o_flag(
        root_resolve_parser,
        "reopen",
        default=None,
        help="reopen the handle with these O_* flags",
    )
    root_resolve_parser.add_argument("subpath", help="path inside the root")
    root_resolve_parser.set_defaults(func=root_resolve)
    del root_resolve_parser

    # root open [--oflags <oflags>] [--[no-]follow] subpath
    root_open_parser = root_subparser.add_parser(
        "open", help="open a path inside the root"
    )
    root_open_parser.add_argument(
        "--follow",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="follow trailing symlinks",
    )
    add_o_flag(
        root_open_parser, "oflags", help="O_* flags to use when opening the file"
    )
    root_open_parser.add_argument("subpath", help="path inside the root")
    root_open_parser.set_defaults(func=root_open)
    del root_open_parser

    # root mkfile [--oflags <oflags>] [--mode <mode>] subpath
    root_mkfile_parser = root_subparser.add_parser(
        "mkfile", help="make an empty file inside the root"
    )
    add_o_flag(
        root_mkfile_parser, "oflags", help="O_* flags to use when creating the file"
    )
    add_mode_flag(root_mkfile_parser, "mode", help="created file mode")
    root_mkfile_parser.add_argument("subpath", help="path inside the root")
    root_mkfile_parser.set_defaults(func=root_mkfile)
    del root_mkfile_parser

    # root mkdir [--mode <mode>] subpath
    root_mkdir_parser = root_subparser.add_parser(
        "mkdir", help="make an empty directory inside the root"
    )
    add_mode_flag(
        root_mkdir_parser,
        "mode",
        default=0o755,
        help="file mode for the created directory",
    )
    root_mkdir_parser.add_argument("subpath", help="path inside the root")
    root_mkdir_parser.set_defaults(func=root_mkdir)
    del root_mkdir_parser

    # root mkdir-all [--mode <mode>] subpath
    root_mkdir_all_parser = root_subparser.add_parser(
        "mkdir-all",
        help="make a directory (including parents) inside the root",
    )
    add_mode_flag(
        root_mkdir_all_parser,
        "mode",
        default=0o755,
        help="file mode for the created directories",
    )
    root_mkdir_all_parser.add_argument("subpath", help="path inside the root")
    root_mkdir_all_parser.set_defaults(func=root_mkdir_all)
    del root_mkdir_all_parser

    # root mknod [--mode <mode>] <subpath> <type> [<major> <minor>]
    root_mknod_parser = root_subparser.add_parser(
        "mknod", help="make an inode inside the root"
    )
    add_mode_flag(root_mknod_parser, "mode", help="file mode for the new inode")
    root_mknod_parser.add_argument("subpath", help="path inside the root")

    class MknodTypeAction(argparse.Action):
        def __call__(
            self: Self,
            parser: argparse.ArgumentParser,
            namespace: argparse.Namespace,
            values: Optional[str | Sequence[str]],
            option_string: Optional[str] = None,
        ):
            inode_type: str
            dev: int
            match values:
                case (inode_type, major, minor):
                    dev = os.makedev(int(major), int(minor))
                case (inode_type,):
                    dev = 0
                case _:
                    raise ValueError(f"invalid mknod type {values}")
            try:
                inode_ftype = {
                    "f": stat.S_IFREG,
                    "d": stat.S_IFDIR,
                    "p": stat.S_IFIFO,
                    "b": stat.S_IFBLK,
                    "c": stat.S_IFCHR,
                    "u": stat.S_IFCHR,  # alias for "c"
                }[inode_type]
            except KeyError:
                raise ValueError(f"invalid mknod type {values}")
            setattr(namespace, self.dest, (inode_ftype, dev))

    root_mknod_parser.add_argument(
        "type",
        nargs=argparse.PARSER,
        action=MknodTypeAction,
        help="inode type to create (like mknod(1))",
    )
    root_mknod_parser.set_defaults(func=root_mknod)
    del root_mknod_parser

    # root hardlink <target> <linkname>
    root_hardlink_parser = root_subparser.add_parser(
        "hardlink", help="make a hardlink inside the root"
    )
    root_hardlink_parser.add_argument(
        "target",
        help="target path of the hardlink inside the root (must already exist)",
    )
    root_hardlink_parser.add_argument(
        "linkname", help="path inside the root for the new hardlink"
    )
    root_hardlink_parser.set_defaults(func=root_hardlink)
    del root_hardlink_parser

    # root symlink <target> <linkname>
    root_symlink_parser = root_subparser.add_parser(
        "symlink", help="make a symbolic link inside the root"
    )
    root_symlink_parser.add_argument("target", help="target path of the symlink")
    root_symlink_parser.add_argument(
        "linkname", help="path inside the root for the new symlink"
    )
    root_symlink_parser.set_defaults(func=root_symlink)
    del root_symlink_parser

    # root readlink <subpath>
    root_readlink_parser = root_subparser.add_parser(
        "readlink", help="read the target path of a symbolic link inside the root"
    )
    root_readlink_parser.add_argument("subpath", help="path inside the root")
    root_readlink_parser.set_defaults(func=root_readlink)
    del root_readlink_parser

    # root unlink <subpath>
    root_unlink_parser = root_subparser.add_parser(
        "unlink", help="remove a file inside the root"
    )
    root_unlink_parser.add_argument("subpath", help="path inside the root")
    root_unlink_parser.set_defaults(func=root_unlink)
    del root_unlink_parser

    # root rmdir <subpath>
    root_rmdir_parser = root_subparser.add_parser(
        "rmdir", help="remove an (empty) directory inside the root"
    )
    root_rmdir_parser.add_argument("subpath", help="path inside the root")
    root_rmdir_parser.set_defaults(func=root_rmdir)
    del root_rmdir_parser

    # root rmdir-all <subpath>
    root_rmdir_all_parser = root_subparser.add_parser(
        "rmdir-all", help="remove a path (recursively) inside the root"
    )
    root_rmdir_all_parser.add_argument("subpath", help="path inside the root")
    root_rmdir_all_parser.set_defaults(func=root_rmdir_all)
    del root_rmdir_all_parser

    # root rename [--exchange] [--whiteout] [--[no-]clobber] <src> <dst>
    root_rename_parser = root_subparser.add_parser(
        "rename", help="rename a path inside the root"
    )
    root_rename_parser.add_argument(
        "--whiteout",
        action="store_true",
        help="create whiteout inode in place of source",
    )
    root_rename_parser.add_argument(
        "--exchange",
        action="store_true",
        help="swap source and destination inodes",
    )
    root_rename_parser.add_argument(
        "--clobber",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="allow rename target to be clobbered",
    )
    root_rename_parser.add_argument("source", help="source path inside the root")
    root_rename_parser.add_argument(
        "destination", help="destination path inside the root"
    )
    root_rename_parser.set_defaults(func=root_rename)
    del root_rename_parser

    del root_subparser, root_parser

    def parse_procfs_base(base: str) -> ProcfsBase:
        if base.startswith("pid="):
            pid = int(base.removeprefix("pid="))
            return procfs.PROC_PID(pid)
        else:
            match base:
                case "root":
                    return procfs.PROC_ROOT
                case "self":
                    return procfs.PROC_SELF
                case "thread-self":
                    return procfs.PROC_THREAD_SELF
                case _:
                    raise ValueError(f"invalid procfs base {base:r}")

    # procfs [--unmasked] --base <base> ... commands
    procfs_parser = top_subparser.add_parser(
        "procfs", help="ProcfsHandle::* operations"
    )
    procfs_parser.add_argument(
        "--unmasked", action="store_true", help="use unmasked procfs handle"
    )
    procfs_parser.add_argument(
        "--base",
        dest="procfs_base",
        type=parse_procfs_base,
        metavar="PROC_*",
        default=procfs.PROC_ROOT,
        help="base path for procfs operations (root, pid=<n>, self, thread-self)",
    )
    procfs_subparser = procfs_parser.add_subparsers()

    # procfs open [--oflags <oflags>] [--[no-]follow] subpath
    procfs_open_parser = procfs_subparser.add_parser(
        "open", help="open a subpath in procfs"
    )
    add_o_flag(
        procfs_open_parser, "oflags", help="O_* flags to use when opening the file"
    )
    procfs_open_parser.add_argument(
        "--follow",
        default=True,
        action=argparse.BooleanOptionalAction,
        help="follow trailing symlinks",
    )
    procfs_open_parser.add_argument("subpath", help="path inside procfs base")
    procfs_open_parser.set_defaults(func=procfs_open)
    del procfs_open_parser

    # procfs readlink <subpath>
    procfs_readlink_parser = procfs_subparser.add_parser(
        "readlink", help="read the target path of a symbolic link in procfs"
    )
    procfs_readlink_parser.add_argument("subpath", help="path inside procfs base")
    procfs_readlink_parser.set_defaults(func=procfs_readlink)
    del procfs_readlink_parser

    del procfs_subparser, procfs_parser

    return parser, parser.parse_args(args)


def main(*argv: str):
    if os.environ.get("PR_SET_DUMPABLE", "") != "":
        raise NotImplementedError(
            "PR_SET_DUMPABLE is not supported by the python pathrs-cmd implementation"
        )

    parser, args = parse_args(argv)
    if args.func is not None:
        args.func(args)
    else:
        # Default to help page if no subcommand was selected.
        parser.print_help()


if __name__ == "__main__":
    try:
        main(*sys.argv[1:])
    except pathrs.PathrsError as e:
        errstr = "<unknown error>"
        if e.errno is not None:
            errstr = os.strerror(e.errno)
        print(f"ERRNO {e.errno} ({errstr})")
        print(f"error: {e.message}")
        sys.exit(1)
